/*
 * Win32 kernel time functions
 *
 * Copyright 1995 Martin von Loewis and Cameron Heide
 * Copyright 2002,2005 CodeWeavers, Aric Stewart
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "config.h"

#include <string.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif
#ifdef HAVE_SYS_TIMES_H
# include <sys/times.h>
#endif
#ifdef HAVE_SYS_LIMITS_H
#include <sys/limits.h>
#elif defined(HAVE_MACHINE_LIMITS_H)
#include <machine/limits.h>
#else
#include <limits.h>
#endif

#ifdef linux
#include <endian.h>
#endif

#if defined linux || defined __APPLE__
#include "tzfile.h"
#endif

#ifdef __APPLE__
#include <CoreFoundation/CFTimeZone.h>
#include <CoreFoundation/CFString.h>
#define fread_unlocked fread
#endif

#define NONAMELESSUNION
#define NONAMELESSSTRUCT
#include "ntstatus.h"
#define WIN32_NO_STATUS
#include "windef.h"
#include "winbase.h"
#include "winternl.h"
#include "kernel_private.h"
#include "wine/unicode.h"
#include "wine/debug.h"

WINE_DEFAULT_DEBUG_CHANNEL(time);

/* maximum time adjustment in seconds for SetLocalTime and SetSystemTime */
#define SETTIME_MAX_ADJUST 120
#define CALINFO_MAX_YEAR 2029

#define LL2FILETIME( ll, pft )\
    (pft)->dwLowDateTime = (UINT)(ll); \
    (pft)->dwHighDateTime = (UINT)((ll) >> 32);
#define FILETIME2LL( pft, ll) \
    ll = (((LONGLONG)((pft)->dwHighDateTime))<<32) + (pft)-> dwLowDateTime ;


static const int MonthLengths[2][12] =
{
	{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
	{ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
};

/* 
 * This structure is used to store strings that represent all of the time zones
 *  in the world. (This is used to help GetTimeZoneInformation)
 */
struct tagTZ_INFO
{
    /*
     * These two are from the unix /usr/share/zoneinfo files
     */
    const char *psTZFromUnix0;
    const char *psTZFromUnix1;
    /*
     * this is the windows registry key for this timezone
     */
    CHAR psTZWindows[32];
    /*
     * these are also from the return zoneinfo are are use to help find a
     * difinitive match
     */
    int bias;
    int dst;
    /* 0 if no possiblility for dups, otherwise set based on number of
     * transisionts this timezone has seen. This should help with things like
     * Central American vs Central Time US vs Mexico Standard Time
     */
    int types;
    const char *psTZFromUnixFirst;
};

/*
 * These are searched top to bottom with the first match being returned
 * There are a number of duplicates which we will need to work harder
 * To identify uniquly.
 * Please feel free to add more if you find windows registry values for more
 * time zones. */

static const struct tagTZ_INFO TZ_INFO[] =
{
    {"MHT","MHT","Dateline Standard Time",-43200,0,0,""},
    {"SST","SST","Samoa Standard Time",39600,0,0,""},
    {"HST","HPT","Hawaiian Standard Time",36000,1,0,""},
    {"AKST","AKDT","Alaskan Standard Time", 32400,1,0,""},
    {"PST","PDT","Pacific Standard Time",28800,1,0,""},
    {"MST","MDT","Mountain Standard Time",25200,1,4,""}, 
    {"MST","MDT","Mexico Standard Time 2",25200,1,5,""},
    {"MST","MDT","US Mountain Standard Time",25200,1,3,""},
    /*catch*/{"MST","MDT","Mountain Standard Time",25200,1,0,""}, 
    {"CST","CDT","Central Standard Time",21600,1,5,"CDT"},
    {"CST","MDT","Canada Central Standard Time",21600,0,0,""},
    {"CST","CDT","Mexico Standard Time",21600,1,5,"LMT"},
    {"CST","CDT","Central America Standard Time",21600,1,3,""}, 
    /*catch*/{"CST","CDT","Central Standard Time",21600,1,0,""},
    {"EST","EDT","Eastern Standard Time",18000,1,0,""},
    {"EST","CDT","US Eastern Standard Time",18000,0,0,""},
    {"COT","COST","SA Pacific Standard Time",18000,1,0,""},/*Bogota*/
    {"PET","PEST","SA Pacific Standard Time",18000,1,0,""},
    {"AST","ADT","Atlantic Standard Time",14400,1,0,""},
    {"VET","VET","SA Western Standard Time",14400,0,0,""},
    {"CLT","CLST","Pacific SA Standard Time",14400,1,0,""},
    {"NST","NDT","Newfoundland Standard Time",12600,1,0,""},
    {"BRT","BRST","E. South America Standard Time",10800,1,0,""},
    {"ART","ARST","SA Eastern Standard Time",10800,0,0,""},
    /*Greenland Standard Time*/
    /* Mid-Atlantic Standard Time */
    {"AZOT","AZOST","Azores Standard Time",3600,1,0,""},
    {"CVT","CVST","Cape Verde Standard Time",3600,0,0,""},
    {"GMT","BST","GMT Standard Time",0,1,0,""},
    {"WET","WEST","Greenwich Standard Time",0,1,0,""},
    {"CET","CEST","Central Europe Standard Time",-3600,1,6,""},
    {"CET","CEST","Central European Standard Time",-3600,1,10,""},
    {"CET","CEST","Romance Standard Time",-3600,1,12,""},
    /*catch*/ {"CET","CEST","Central Europe Standard Time",-3600,1,0,""},
    {"WET","WEST","W. Europe Standard Time",-3600,1,0,""}, 
    /* W. Central Africa Standard Time */
    {"EET","EEST","E. Europe Standard Time",-7200,1,7,""},
    {"EET","EEST","Egypt Standard Time",-7200,1,4,""}, 
    {"EET","EEST","FLE Standard Time",-7200,1,5,""},
    {"EET","EEST","GTB Standard Time",-7200,1,9,""},
    /*catch*/{"EET","EEST","E. Europe Standard Time",-7200,1,0,""},
    {"IST","IDT","Israel Standard Time",-7200,1,0,""},
    {"SAST","SAST","South Africa Standard Time",-7200,1,0,""},
    {"MSK","MSD","Russian Standard Time",-10800,1,0,""},
    {"AST","AST","Arab Standard Time",-10800,0,0,""},
    {"EAT","EAT","E. Africa Standard Time",-10800,0,0,""},
    {"AST","ADT","Arabic Standard Time",-10800,1,0,""},
    {"IRST","IRDT","Iran Standard Time",-12600,1,0,""},
    {"GST","GST","Arabian Standard Time",-14400,0,0,""},
    {"AZT","AZST","Caucasus Standard Time",-14400,1,0,""},
    {"AFT","AFT", "Afghanistan Standard Time", -16200,0,0,""},
    /* Ekaterinburg Standard Time */
    {"PKT","PKST","West Asia Standard Time",-18000,1,0,""},
    {"IST","IST","India Standard Time",-19800,1,0,""},
    /* Nepal Standard Time */
    {"BDT","BDT","Central Asia Standard Time",-21600,0,0,""},
    /* Sri Lanka Standard Time */
    {"NOVT","NOVST","N. Central Asia Standard Time", -21600,1,0,""},
    {"MMT","MMT","Myanmar Standard Time",-23400,0,0,""},
    {"ICT","ICT","SE Asia Standard Time",-25200,0,0,""},
    {"KRAT","KRAST","North Asia Standard Time",-25200,1,0,""},
    {"CST","CDT","China Standard Time",-28800,1,4,""},
    {"SGT","MALST","Singapore Standard Time",-28800,1,0,""},
    {"CST","CDT","Taipei Standard Time",-28800,1,2,""},
    /*catch*/{"CST","CDT","China Standard Time",-28800,1,0,""},
    {"WST","WST","W. Australia Standard Time",-28800,1,0,""},
    {"IRKT","IRKST","North Asia East Standard Time",-28800,1,0,""},
    {"KST","KDT","Korea Standard Time",-32400,1,0,""},
    {"JST","JST","Tokyo Standard Time",-32400,0,0,""},
    {"YAKT","YAKST","Yakutsk Standard Time",-32400,1,0,""},
    {"CST","CST","AUS Central Standard Time",-34200,1,2,""},
    /*catch*/{"CST","CST","Cen. Australia Standard Time",-34200,1,0,""}, 
    /*AUS Eastern Standard Time*/
    {"EST","EST","E. Australia Standard Time",-36000,1,0,""},
    /*Tasmania Standard Time*/
    {"VLAT","VLAST","Vladivostok Standard Time",-36000,1,0,""},
    {"ChST","ChST","West Pacific Standard Time", -36000,0,0,""},
    {"MAGT","MAGST","Central Pacific Standard Time",-39600,1,0,""},
    {"FJT","FJST","Fiji Standard Time",-43200,1,0,""},
    {"NZST","NZDT","New Zealand Standard Time",-43200,1,0,""},
    {"TOT","TOST","Tonga Standard Time",-46800,1,0,""},
};


/****************************************************************************
 * The following code was lifted from glibc under the LGPL and modified to do
 * what we need it to do.
 ****************************************************************************/

#if defined linux || defined __APPLE__

/* Here are the main values! */
static char *wine__tzname[2];
static char *wine__tzcode;
static int wine__daylight;
static long int wine__timezone;

/******
 * STRUCTS
 ******/
struct ttinfo
  {
    long int offset;        /* Seconds east of GMT.  */
    unsigned char isdst;    /* Used to set tm_isdst.  */
    unsigned char idx;      /* Index into `zone_names'.  */
    unsigned char isstd;    /* Transition times are in standard time.  */
    unsigned char isgmt;    /* Transition times are in GMT.  */
  };

struct leap
  {
    time_t transition;      /* Time the transition takes effect.  */
    long int change;        /* Seconds of correction to apply.  */
  };

/* List of buffers containing time zone strings. */
struct tzstring_l
{
  struct tzstring_l *next;
  size_t len;  /* strlen(data) - doesn't count terminating NUL! */
  char data[0];
};

/*********
 * GLOBALS
 *********/

struct tzstring_l *tzstring_list;
static size_t num_transitions=0;
static time_t *transitions=NULL;
static unsigned char *type_idxs=NULL;
static size_t num_types=0;
static struct ttinfo *types=NULL;
static char *zone_names=NULL;
static long int rule_stdoff=0;
static long int rule_dstoff=0;
static size_t num_leaps=0;
static struct leap *leaps=NULL;
static int __use_tzfile=0;
static size_t __tzname_cur_max = 0;

/************
 * FUNCTIONS
 ************/

static void compute_tzname_max (size_t chars)
{
  const char *p;

  p = zone_names;
  do
    {
      const char *start = p;
      while (*p != '\0')
    ++p;
      if ((size_t) (p - start) > __tzname_cur_max)
    __tzname_cur_max = p - start;
    }
  while (++p < &zone_names[chars]);
}

/* Decode the four bytes at PTR as a signed integer in network byte order.  */
static inline int decode (const void *ptr)
{
  if ((BYTE_ORDER == BIG_ENDIAN) && sizeof (int) == 4)
    return *(const int *) ptr;
  else if (BYTE_ORDER == LITTLE_ENDIAN && sizeof (int) == 4)
    return RtlUlongByteSwap (*(const int *) ptr);
  else
    {
      const unsigned char *p = ptr;
      int result = *p & (1 << (CHAR_BIT - 1)) ? ~0 : 0;

      result = (result << 8) | *p++;
      result = (result << 8) | *p++;
      result = (result << 8) | *p++;
      result = (result << 8) | *p++;

      return result;
    }
}

char * __tzstring (const char *s)
{
  char *p;
  struct tzstring_l *t, *u, *new;
  size_t len = strlen(s);

  /* Walk the list and look for a match.  If this string is the same
     as the end of an already-allocated string, it can share space. */
  for (u = t = tzstring_list; t; u = t, t = t->next)
    if (len <= t->len)
      {
    p = &t->data[t->len - len];
    if (strcmp (s, p) == 0)
      return p;
      }

  /* Not found; allocate a new buffer. */
  new = malloc (sizeof (struct tzstring_l) + len + 1);
  if (!new)
    return NULL;

  new->next = NULL;
  new->len = len;
  strcpy (new->data, s);

  if (u)
    u->next = new;
  else
    tzstring_list = new;

  return new->data;
}

void __tzfile_read (const char *file, size_t extra, char **extrap)
{
  static const char default_tzdir[] = TZDIR;
  size_t num_isstd, num_isgmt;
  register FILE *f;
  struct tzhead tzhead;
  size_t chars;
  register size_t i;
  size_t total_size;
  size_t types_idx;
  size_t leaps_idx;
 
  __use_tzfile = 0; 

  if (transitions != NULL)
    free ((void *) transitions);

  transitions = NULL;

  if (file == NULL)
{
    /* No user specification; use the site-wide default.  */
    file = TZDEFAULT;
}
  else if (*file == '\0')
  {
    /* User specified the empty string; use UTC with no leap seconds.  */
    return;
  }
  else
    {
      /* We must not allow to read an arbitrary file in a setuid
     program.  So we fail for any file which is not in the
     directory hierachy starting at TZDIR
     and which is not the system wide default TZDEFAULT.  */
      if (( *file == '/'
           && memcmp (file, TZDEFAULT, sizeof(TZDEFAULT))
           && memcmp (file, default_tzdir, sizeof (default_tzdir) - 1))
          || strstr (file, "../") != NULL)
    {
        /* This test is certainly a bit too restrictive but it should
       catch all critical cases.  */
	ERR("security check failed, check location of localtime file\n");
        return;
    }
    }

  if (*file != '/')
    {
      const char *tzdir;
      unsigned int len, tzdir_len;
      char *new, *tmp;

      tzdir = getenv ("TZDIR");
      if (tzdir == NULL || *tzdir == '\0')
    {
      tzdir = default_tzdir;
      tzdir_len = strlen (default_tzdir);
    }
      else
    tzdir_len = strlen (tzdir);

      len = strlen (file) + 1;
      new = (char *) malloc (tzdir_len + 1 + len);
      memcpy (new, tzdir, tzdir_len);
      tmp = new+tzdir_len;
      *tmp++ = '/';
      memcpy (tmp, file, len);
      file = new;
    }

  f = fopen (file, "r");
  if (f == NULL)
    return;

  if (fread_unlocked ((void *) &tzhead, sizeof (tzhead), 1, f) != 1)
    goto lose;

  num_transitions = (size_t) decode (tzhead.tzh_timecnt);
  num_types = (size_t) decode (tzhead.tzh_typecnt);
  chars = (size_t) decode (tzhead.tzh_charcnt);
  num_leaps = (size_t) decode (tzhead.tzh_leapcnt);
  num_isstd = (size_t) decode (tzhead.tzh_ttisstdcnt);
  num_isgmt = (size_t) decode (tzhead.tzh_ttisgmtcnt);

  total_size = num_transitions * (sizeof (time_t) + 1);
  total_size = ((total_size + __alignof__ (struct ttinfo) - 1)
        & ~(__alignof__ (struct ttinfo) - 1));
  types_idx = total_size;
  total_size += num_types * sizeof (struct ttinfo) + chars;
  total_size = ((total_size + __alignof__ (struct leap) - 1)
        & ~(__alignof__ (struct leap) - 1));
  leaps_idx = total_size;
  total_size += num_leaps * sizeof (struct leap);
  /* This is for the extra memory required by the caller.  */
  total_size += extra;

  transitions = (time_t *) malloc (total_size);
  if (transitions == NULL)
    goto lose;

  type_idxs = (unsigned char *) transitions + (num_transitions
                           * sizeof (time_t));
  types = (struct ttinfo *) ((char *) transitions + types_idx);
  zone_names = (char *) types + num_types * sizeof (struct ttinfo);
  leaps = (struct leap *) ((char *) transitions + leaps_idx);
  if (extra > 0)
    *extrap = (char *) &leaps[num_leaps];

  if (sizeof (time_t) < 4)
    abort ();

  if (sizeof (time_t) == 4)
    {
      if (fread_unlocked (transitions, 1, (4 + 1) * num_transitions, f)
      != (4 + 1) * num_transitions)
    goto lose;
    }
  else
    {
      if (fread_unlocked (transitions, 4, num_transitions, f)
      != num_transitions
      || fread_unlocked (type_idxs, 1, num_transitions, f)
      != num_transitions)
    goto lose;
    }

  /* Check for bogus indices in the data file, so we can hereafter
     safely use type_idxs[T] as indices into `types' and never crash.  */
  for (i = 0; i < num_transitions; ++i)
    if (type_idxs[i] >= num_types)
      goto lose;

  if (BYTE_ORDER != BIG_ENDIAN || sizeof (time_t) != 4)
    {
      /* Decode the transition times, stored as 4-byte integers in
     network (big-endian) byte order.  We work from the end of
     the array so as not to clobber the next element to be
     processed when sizeof (time_t) > 4.  */
      i = num_transitions;
      while (i-- > 0)
    transitions[i] = decode ((char *) transitions + i * 4);
    }

  for (i = 0; i < num_types; ++i)
    {
      unsigned char x[4];
      int c;
      if (fread_unlocked (x, 1, sizeof (x), f) != sizeof (x))
    goto lose;
      c = getc_unlocked (f);
      if ((unsigned int) c > 1u)
    goto lose;
      types[i].isdst = c;
      c = getc_unlocked (f);
      if ((size_t) c > chars) /* Bogus index in data file.  */
    goto lose;
      types[i].idx = c;
      types[i].offset = (long int) decode (x);
    }

  if (fread_unlocked (zone_names, 1, chars, f) != chars)
    goto lose;

  for (i = 0; i < num_leaps; ++i)
    {
      unsigned char x[4];
      if (fread_unlocked (x, 1, sizeof (x), f) != sizeof (x))
    goto lose;
      leaps[i].transition = (time_t) decode (x);
      if (fread_unlocked (x, 1, sizeof (x), f) != sizeof (x))
    goto lose;
      leaps[i].change = (long int) decode (x);
    }

  for (i = 0; i < num_isstd; ++i)
    {
      int c = getc_unlocked (f);
      if (c == EOF)
    goto lose;
      types[i].isstd = c != 0;
    }
  while (i < num_types)
    types[i++].isstd = 0;

  for (i = 0; i < num_isgmt; ++i)
    {
      int c = getc_unlocked (f);
      if (c == EOF)
        goto lose;
      types[i].isgmt = c != 0;
    }
  while (i < num_types)
    types[i++].isgmt = 0;

  fclose (f);

  /* First "register" all timezone names.  */
  for (i = 0; i < num_types; ++i)
  {
    TRACE("%s  (%i %i %i)\n",&zone_names[types[i].idx],types[i].isdst,
                            types[i].isstd, types[i].isgmt);

    (void) __tzstring (&zone_names[types[i].idx]);

  }


  /* Find the standard and daylight time offsets used by the rule file.
     We choose the offsets in the types of each flavor that are
     transitioned to earliest in time.  */
  wine__tzname[0] = NULL;
  wine__tzname[1] = NULL;
  wine__tzcode = NULL;

  /* for dup matching */
  wine__tzcode = __tzstring(&zone_names[types[0].idx]);;

  for (i = num_transitions; i > 0; )
    {
      int type = type_idxs[--i];
      int dst = types[type].isdst;

      if (wine__tzname[dst] == NULL)
    {
      int idx = types[type].idx;

      wine__tzname[dst] = __tzstring (&zone_names[idx]);

      if (wine__tzname[1 - dst] != NULL)
        break;
    }
    }
  if (wine__tzname[0] == NULL)
    {
      /* This should only happen if there are no transition rules.
     In this case there should be only one single type.  */
      wine__tzname[0] = __tzstring (zone_names);
    }
  if (wine__tzname[1] == NULL)
    wine__tzname[1] = wine__tzname[0];

  compute_tzname_max (chars);

  if (num_transitions == 0)
    /* Use the first rule (which should also be the only one).  */
    rule_stdoff = rule_dstoff = types[0].offset;
  else
    {
      int stdoff_set = 0, dstoff_set = 0;
      rule_stdoff = rule_dstoff = 0;
      i = num_transitions - 1;
      do
    {
      if (!stdoff_set && !types[type_idxs[i]].isdst)
        {
          stdoff_set = 1;
          rule_stdoff = types[type_idxs[i]].offset;
        }
      else if (!dstoff_set && types[type_idxs[i]].isdst)
        {
          dstoff_set = 1;
          rule_dstoff = types[type_idxs[i]].offset;
        }
      if (stdoff_set && dstoff_set)
        break;
    }
      while (i-- > 0);

      if (!dstoff_set)
    rule_dstoff = rule_stdoff;
    }

  TRACE("stdoff %lx, dstoff %lx\n",rule_stdoff,rule_dstoff);

  /* does daylight savings time ever apply in this timezone? NB: NOT does DST apply right now */
  wine__daylight = rule_stdoff != rule_dstoff;
  wine__timezone = -rule_stdoff;

  __use_tzfile = 1;
  return;

 lose:
  fclose (f);
}

static void TIME_GetTZTransitionDates( time_t *dst, time_t *sdt)
{
    time_t currtime;
    int i,j;

    *dst = 0;
    *sdt = 0;
    
    if (!wine__daylight)
        return;

    time(&currtime);

    for (i = 1; i < num_transitions; ++i)
        if (currtime < transitions[i])
            break;

    if (i == num_transitions)
        return;

    j = type_idxs[i-1];  

    if (types[j].isdst)
    {
        *dst = transitions[i-1];
        *sdt = transitions[i];
    }
    else 
    {
        j = type_idxs[i];
        if (types[j].isdst) 
        {
            *sdt = transitions[i-1];
            *dst = transitions[i];
        }  
    }
}

static void TIME_InitTZInfo(void)
{
  register const char *tz;
  static char *old_tz = NULL;
  long timezone_bias;

#ifdef __APPLE__
  CFTimeZoneRef tz_ref;
  CFStringRef tz_name_str_ref;
  char tz_name[256];

  tz_ref = CFTimeZoneCopySystem();
  tz_name_str_ref = CFTimeZoneGetName( tz_ref );
  CFStringGetCString( tz_name_str_ref, tz_name, sizeof(tz_name), kCFStringEncodingUTF8 );
  tz = tz_name;
  TRACE("tz = %s\n", tz);
  timezone_bias = CFTimeZoneGetSecondsFromGMT( tz_ref, CFAbsoluteTimeGetCurrent() );
  CFRelease( tz_ref );
#else
  /* Examine the TZ environment variable.  */
  tz = getenv ("TZ");
  if (tz == NULL)
    /* No user specification; use the site-wide default.  */
    tz = TZDEFAULT;
  else if (*tz == '\0')
    /* User specified the empty string; use UTC explicitly.  */
    tz = "Universal";
  /* A leading colon means "implementation defined syntax".
     We ignore the colon and always use the same algorithm:
     try a data file, and if none exists parse the 1003.1 syntax.  */
  if (tz && *tz == ':')
    ++tz;
  timezone_bias = timezone;
#endif

  /* Check whether the value changes since the last run.  */
  if (old_tz != NULL && tz != NULL && strcmp (tz, old_tz) == 0)
    /* No change, simply return.  */
    return;

  /* Save the value of `tz'.  */
  if (old_tz != NULL)
    free (old_tz);
  old_tz = tz ? strdup (tz) : NULL;

  /* Try to read a data file.  */
  __tzfile_read (tz, 0, NULL);

  if (!__use_tzfile)
  {
      tzset();
      wine__tzname[0] = tzname[0];
      wine__tzname[1] = tzname[1];
      wine__tzcode = NULL;
      wine__timezone = timezone_bias;
      wine__daylight = daylight;
  } 
}

static DWORD TIME_GetTimeNameFromReg (const CHAR* key, const CHAR* type, 
                                      WCHAR* name, INT size)
{
    HKEY hkey;
    LONG rc;
    OBJECT_ATTRIBUTES attr;
    UNICODE_STRING nameW;
    CHAR  skeya[1024];
    WCHAR subkey[1024];
    WCHAR skey[9];

    strcpy(skeya,"Machine\\Software\\Microsoft\\Windows\\CurrentVersion\\Time Zones\\");
    strcat(skeya,key);

    MultiByteToWideChar(CP_ACP,0, skeya, -1,subkey,1024);
    MultiByteToWideChar(CP_ACP,0, type, -1,skey, 9);

    attr.Length = sizeof(attr);
    attr.RootDirectory = 0;
    attr.ObjectName = &nameW;
    attr.Attributes = 0;
    attr.SecurityDescriptor = NULL;
    attr.SecurityQualityOfService = NULL;
    RtlInitUnicodeString( &nameW, subkey );
    rc = NtOpenKey( (PHANDLE)&hkey, KEY_ALL_ACCESS, &attr );

    if (rc == 0)
    {
        char buffer[1024];
        KEY_VALUE_PARTIAL_INFORMATION *info = (KEY_VALUE_PARTIAL_INFORMATION *)buffer;
        DWORD total_size;

        RtlInitUnicodeString( &nameW, skey );
        rc = NtQueryValueKey( hkey, &nameW, KeyValuePartialInformation,
                              buffer, sizeof(buffer), &total_size );
        if (!rc)
        {
            INT len = min( info->DataLength, (size-1) * sizeof(WCHAR) );
            memcpy( name, info->Data, len );
            name[len / sizeof(WCHAR)] = 0;
            TRACE("returning %s\n",debugstr_w(name));
        }
        NtClose(hkey);
    }
    if (rc)
    {
        MultiByteToWideChar(CP_ACP,0, key, -1,name,size);
    }
    
    return rc;
}

/* TIME_GetTZInfo: helper function that returns the given timezone as a string.
   This could be done with a hash table instead of merely iterating through
   a table, however with the small amount of entries (60 or so) I didn't think
   it was worth it.
   
   This function should return a status code of some sort, for if the requested
   timezone isn't found.
*/

static VOID TIME_GetWindowsTZInfo (WCHAR *StandardName, 
                                           WCHAR *DaylightName, 
                                           SYSTEMTIME *DaylightDate,
                                           SYSTEMTIME *StandardDate)
{
   int i;

   TRACE("%s  %s %i %li (%i)\n",wine__tzname[0],wine__tzname[1],wine__daylight,wine__timezone,(int)num_types);

   TIME_InitTZInfo();

   memset(DaylightDate,0,sizeof(SYSTEMTIME));
   memset(StandardDate,0,sizeof(SYSTEMTIME));

   for (i=0; i<(sizeof(TZ_INFO) / sizeof(struct tagTZ_INFO)); i++)
   {
      if ( strcmp(TZ_INFO[i].psTZFromUnix0, wine__tzname[0]) == 0 &&
           (strcmp(TZ_INFO[i].psTZFromUnix1, wine__tzname[1]) == 0) &&
           TZ_INFO[i].bias == wine__timezone &&
           TZ_INFO[i].dst == wine__daylight 
         )
        {
            time_t dst, sdt;
            struct tm *p;

            /* check for dups */
            if (TZ_INFO[i].types && TZ_INFO[i].types != num_types)
                continue;
            /* next check for dups */
            if (TZ_INFO[i].psTZFromUnixFirst[0] &&
                    strcmp(TZ_INFO[i].psTZFromUnixFirst,wine__tzcode)!=0)
                continue;

            TIME_GetTimeNameFromReg (TZ_INFO[i].psTZWindows, "Std", 
                                     StandardName, 32);
            TIME_GetTimeNameFromReg (TZ_INFO[i].psTZWindows, "Dlt", 
                                     DaylightName, 32);
      
            TIME_GetTZTransitionDates(&dst,&sdt);

            if (dst && sdt)
            {
                p = localtime(&dst);

		/* the notation for these dates is somewhat odd, see the MSDN documentation on TIME_ZONE_INFORMATION */
                DaylightDate->wYear = 0;
                DaylightDate->wMonth = p->tm_mon+1;
                DaylightDate->wDay = (p->tm_mday - 1) / 7 + 1;
                DaylightDate->wHour = p->tm_hour;
                DaylightDate->wMinute = 0;
                DaylightDate->wSecond = 0;
		DaylightDate->wDayOfWeek = p->tm_wday;

                p = localtime(&sdt);

                StandardDate->wYear = 0;
                StandardDate->wMonth = p->tm_mon+1;
                StandardDate->wDay = (p->tm_mday - 1) / 7 + 1;
                StandardDate->wHour = 0;
                StandardDate->wMinute = 0;
                StandardDate->wSecond = 0;
		StandardDate->wDayOfWeek = p->tm_wday;

		TRACE("DaylightDate=dd/mm/yy %02d/%02d/%02d\n", DaylightDate->wYear, DaylightDate->wMonth, DaylightDate->wDay);
		TRACE("StandardDate=dd/mm/yy %02d/%02d/%02d\n", StandardDate->wYear, StandardDate->wMonth, StandardDate->wDay);
		
            }
 
            break;     
        }
   }
}

/* returns 1 if we are currently within daylight savings,
           0 if we're not,
	   -1 on error
   given the start dates of both. assumption is std is earlier than daylight
 */
static int check_daylight_savings(void)
{
    time_t now;
    struct tm now_broken;
    time(&now);
    if (!localtime_r(&now, &now_broken)) {
	ERR("localtime_r failed\n");
	return -1;
    }
    return now_broken.tm_isdst;
}
#endif

static inline int IsLeapYear(int Year)
{
	return Year % 4 == 0 && (Year % 100 != 0 || Year % 400 == 0) ? 1 : 0;
}

/***********************************************************************
 *  TIME_DayLightCompareDate
 *
 * Compares two dates without looking at the year.
 *
 * PARAMS
 *   date        [in] The local time to compare.
 *   compareDate [in] The daylight saving begin or end date.
 *
 * RETURNS
 *
 *  -1 if date < compareDate
 *   0 if date == compareDate
 *   1 if date > compareDate
 *  -2 if an error occurs
 */
static int TIME_DayLightCompareDate( const SYSTEMTIME *date,
    const SYSTEMTIME *compareDate )
{
    int limit_day, dayinsecs;

    if (date->wMonth < compareDate->wMonth)
        return -1; /* We are in a month before the date limit. */

    if (date->wMonth > compareDate->wMonth)
        return 1; /* We are in a month after the date limit. */

    if (compareDate->wDayOfWeek <= 6)
    {
        WORD First;
        /* compareDate->wDay is interpreted as number of the week in the month
         * 5 means: the last week in the month */
        int weekofmonth = compareDate->wDay;
          /* calculate the day of the first DayOfWeek in the month */
        First = ( 6 + compareDate->wDayOfWeek - date->wDayOfWeek + date->wDay 
               ) % 7 + 1;
        limit_day = First + 7 * (weekofmonth - 1);
        /* check needed for the 5th weekday of the month */
        if(limit_day > MonthLengths[date->wMonth==2 && IsLeapYear(date->wYear)]
                [date->wMonth - 1])
            limit_day -= 7;
    }
    else
    {
       limit_day = compareDate->wDay;
    }

    /* convert to seconds */
    limit_day = ((limit_day * 24  + compareDate->wHour) * 60 +
            compareDate->wMinute ) * 60;
    dayinsecs = ((date->wDay * 24  + date->wHour) * 60 +
            date->wMinute ) * 60 + date->wSecond;
    /* and compare */
    return dayinsecs < limit_day ? -1 :
           dayinsecs > limit_day ? 1 :
           0;   /* date is equal to the date limit. */
}

/***********************************************************************
 *  TIME_CompTimeZoneID
 *
 *  Computes the local time bias for a given time and time zone.
 *
 *  PARAMS
 *      pTZinfo     [in] The time zone data.
 *      lpFileTime  [in] The system or local time.
 *      islocal     [in] it is local time.
 *
 *  RETURNS
 *      TIME_ZONE_ID_INVALID    An error occurred
 *      TIME_ZONE_ID_UNKNOWN    There are no transition time known
 *      TIME_ZONE_ID_STANDARD   Current time is standard time
 *      TIME_ZONE_ID_DAYLIGHT   Current time is dayligh saving time
 */
static BOOL TIME_CompTimeZoneID ( const TIME_ZONE_INFORMATION *pTZinfo,
    FILETIME *lpFileTime, BOOL islocal )
{
    int ret;
    BOOL beforeStandardDate, afterDaylightDate;
    DWORD retval = TIME_ZONE_ID_INVALID;
    LONGLONG llTime = 0; /* initialized to prevent gcc complaining */
    SYSTEMTIME SysTime;
    FILETIME ftTemp;

    if (pTZinfo->DaylightDate.wMonth != 0)
    {
        if (pTZinfo->StandardDate.wMonth == 0 ||
            pTZinfo->StandardDate.wDay<1 ||
            pTZinfo->StandardDate.wDay>5 ||
            pTZinfo->DaylightDate.wDay<1 ||
            pTZinfo->DaylightDate.wDay>5)
        {
            SetLastError(ERROR_INVALID_PARAMETER);
            return TIME_ZONE_ID_INVALID;
        }

        if (!islocal) {
            FILETIME2LL( lpFileTime, llTime );
            llTime -= ( pTZinfo->Bias + pTZinfo->DaylightBias )
                * (LONGLONG)600000000;
            LL2FILETIME( llTime, &ftTemp)
            lpFileTime = &ftTemp;
        }

        FileTimeToSystemTime(lpFileTime, &SysTime);
        
         /* check for daylight saving */
        ret = TIME_DayLightCompareDate( &SysTime, &pTZinfo->StandardDate);
        if (ret == -2)
          return TIME_ZONE_ID_INVALID;

        beforeStandardDate = ret < 0;

        if (!islocal) {
            llTime -= ( pTZinfo->StandardBias - pTZinfo->DaylightBias )
                * (LONGLONG)600000000;
            LL2FILETIME( llTime, &ftTemp)
            FileTimeToSystemTime(lpFileTime, &SysTime);
        }

        ret = TIME_DayLightCompareDate( &SysTime, &pTZinfo->DaylightDate);
        if (ret == -2)
          return TIME_ZONE_ID_INVALID;

        afterDaylightDate = ret >= 0;

        retval = TIME_ZONE_ID_STANDARD;
        if( pTZinfo->DaylightDate.wMonth <  pTZinfo->StandardDate.wMonth ) {
            /* Northern hemisphere */
            if( beforeStandardDate && afterDaylightDate )
                retval = TIME_ZONE_ID_DAYLIGHT;
        } else    /* Down south */
            if( beforeStandardDate || afterDaylightDate )
            retval = TIME_ZONE_ID_DAYLIGHT;
    } else 
        /* No transition date */
        retval = TIME_ZONE_ID_UNKNOWN;
        
    return retval;
}

#if !defined linux && !defined __APPLE__
/***********************************************************************
 *  TIME_TimeZoneID
 *
 *  Calculates whether daylight saving is on now.
 *
 *  PARAMS
 *      pTzi [in] Timezone info.
 *
 *  RETURNS
 *      TIME_ZONE_ID_INVALID    An error occurred
 *      TIME_ZONE_ID_UNKNOWN    There are no transition time known
 *      TIME_ZONE_ID_STANDARD   Current time is standard time
 *      TIME_ZONE_ID_DAYLIGHT   Current time is dayligh saving time
 */
static DWORD TIME_ZoneID( const TIME_ZONE_INFORMATION *pTzi )
{
    FILETIME ftTime;
    GetSystemTimeAsFileTime( &ftTime);
    return TIME_CompTimeZoneID( pTzi, &ftTime, FALSE);
}
#endif

/***********************************************************************
 *  TIME_GetTimezoneBias
 *
 *  Calculates the local time bias for a given time zone.
 *
 * PARAMS
 *  pTZinfo    [in]  The time zone data.
 *  lpFileTime [in]  The system or local time.
 *  islocal    [in]  It is local time.
 *  pBias      [out] The calculated bias in minutes.
 *
 * RETURNS
 *  TRUE when the time zone bias was calculated.
 */
static BOOL TIME_GetTimezoneBias( const TIME_ZONE_INFORMATION *pTZinfo,
    FILETIME *lpFileTime, BOOL islocal, LONG *pBias )
{
    LONG bias = pTZinfo->Bias;
    DWORD tzid = TIME_CompTimeZoneID( pTZinfo, lpFileTime, islocal);

    if( tzid == TIME_ZONE_ID_INVALID)
        return FALSE;
    if (tzid == TIME_ZONE_ID_DAYLIGHT)
        bias += pTZinfo->DaylightBias;
    else if (tzid == TIME_ZONE_ID_STANDARD)
        bias += pTZinfo->StandardBias;
    *pBias = bias;
    return TRUE;
}


/***********************************************************************
 *              SetLocalTime            (KERNEL32.@)
 *
 *  Set the local time using current time zone and daylight
 *  savings settings.
 *
 * PARAMS
 *  systime [in] The desired local time.
 *
 * RETURNS
 *  Success: TRUE. The time was set.
 *  Failure: FALSE, if the time was invalid or caller does not have
 *           permission to change the time.
 */
BOOL WINAPI SetLocalTime( const SYSTEMTIME *systime )
{
    FILETIME ft;
    LARGE_INTEGER st, st2;
    NTSTATUS status;

    if( !SystemTimeToFileTime( systime, &ft ))
        return FALSE;
    st.u.LowPart = ft.dwLowDateTime;
    st.u.HighPart = ft.dwHighDateTime;
    RtlLocalTimeToSystemTime( &st, &st2 );

    if ((status = NtSetSystemTime(&st2, NULL)))
        SetLastError( RtlNtStatusToDosError(status) );
    return !status;
}


/***********************************************************************
 *           GetSystemTimeAdjustment     (KERNEL32.@)
 *
 *  Get the period between clock interrupts and the amount the clock
 *  is adjusted each interrupt so as to keep it in sync with an external source.
 *
 * PARAMS
 *  lpTimeAdjustment [out] The clock adjustment per interrupt in 100's of nanoseconds.
 *  lpTimeIncrement  [out] The time between clock interrupts in 100's of nanoseconds.
 *  lpTimeAdjustmentDisabled [out] The clock synchronisation has been disabled.
 *
 * RETURNS
 *  TRUE.
 *
 * BUGS
 *  Only the special case of disabled time adjustments is supported.
 */
BOOL WINAPI GetSystemTimeAdjustment( PDWORD lpTimeAdjustment, PDWORD lpTimeIncrement,
    PBOOL  lpTimeAdjustmentDisabled )
{
    *lpTimeAdjustment = 0;
    *lpTimeIncrement = 0;
    *lpTimeAdjustmentDisabled = TRUE;
    return TRUE;
}


/***********************************************************************
 *              SetSystemTime            (KERNEL32.@)
 *
 *  Set the system time in utc.
 *
 * PARAMS
 *  systime [in] The desired system time.
 *
 * RETURNS
 *  Success: TRUE. The time was set.
 *  Failure: FALSE, if the time was invalid or caller does not have
 *           permission to change the time.
 */
BOOL WINAPI SetSystemTime( const SYSTEMTIME *systime )
{
    FILETIME ft;
    LARGE_INTEGER t;
    NTSTATUS status;

    if( !SystemTimeToFileTime( systime, &ft ))
        return FALSE;
    t.u.LowPart = ft.dwLowDateTime;
    t.u.HighPart = ft.dwHighDateTime;
    if ((status = NtSetSystemTime(&t, NULL)))
        SetLastError( RtlNtStatusToDosError(status) );
    return !status;
}

/***********************************************************************
 *              SetSystemTimeAdjustment  (KERNEL32.@)
 *
 *  Enables or disables the timing adjustments to the system's clock.
 *
 * PARAMS
 *  dwTimeAdjustment        [in] Number of units to add per clock interrupt.
 *  bTimeAdjustmentDisabled [in] Adjustment mode.
 *
 * RETURNS
 *  Success: TRUE.
 *  Failure: FALSE.
 */
BOOL WINAPI SetSystemTimeAdjustment( DWORD dwTimeAdjustment, BOOL bTimeAdjustmentDisabled )
{
    /* Fake function for now... */
    FIXME("(%08x,%d): stub !\n", dwTimeAdjustment, bTimeAdjustmentDisabled);
    return TRUE;
}

/***********************************************************************
 *              GetTimeZoneInformation  (KERNEL32.@)
 *
 *  Get information about the current local time zone.
 *
 * PARAMS
 *  tzinfo [out] Destination for time zone information.
 *
 * RETURNS
 *  TIME_ZONE_ID_INVALID    An error occurred
 *  TIME_ZONE_ID_UNKNOWN    There are no transition time known
 *  TIME_ZONE_ID_STANDARD   Current time is standard time
 *  TIME_ZONE_ID_DAYLIGHT   Current time is dayligh saving time
 */
DWORD WINAPI GetTimeZoneInformation( LPTIME_ZONE_INFORMATION tzinfo )
{
#if defined linux || defined __APPLE__
    int ret;
    
    memset(tzinfo, 0, sizeof(TIME_ZONE_INFORMATION));

    TIME_InitTZInfo();

    tzinfo->Bias = wine__timezone / 60;
    tzinfo->StandardBias = 0;
    tzinfo->DaylightBias = -60;

    /* Get the standard time string */
    TIME_GetWindowsTZInfo(tzinfo->StandardName, tzinfo->DaylightName,
                          &(tzinfo->DaylightDate), &(tzinfo->StandardDate));

    TRACE("Zone Names: %s  %s  \n",debugstr_w(tzinfo->StandardName),
                       debugstr_w(tzinfo->DaylightName));

    /* are we currently operating in daylight savings time? */
    ret = check_daylight_savings();
    if (ret == -1) {
	ERR("check_daylight_savings failed, defaulting to no DST\n");
	return TIME_ZONE_ID_UNKNOWN;
    }

    if (ret) TRACE("DST is active\n");
    else TRACE("DST is not active\n");
    
    if (ret)
	return TIME_ZONE_ID_DAYLIGHT;
		  
    /* FIXME: there is a 9x/NT incompatibility here.
     * On NT we should return ID_UNKNOWN and on 9x ID_STANDARD. See MSDN
     */
    return TIME_ZONE_ID_UNKNOWN;
#else
    NTSTATUS status;

    status = RtlQueryTimeZoneInformation( (RTL_TIME_ZONE_INFORMATION*)tzinfo );
    if ( status != STATUS_SUCCESS )
    {
        SetLastError( RtlNtStatusToDosError(status) );
        return TIME_ZONE_ID_INVALID;
    }
    return TIME_ZoneID( tzinfo );
#endif
}

/***********************************************************************
 *              SetTimeZoneInformation  (KERNEL32.@)
 *
 *  Change the settings of the current local time zone.
 *
 * PARAMS
 *  tzinfo [in] The new time zone.
 *
 * RETURNS
 *  Success: TRUE. The time zone was updated with the settings from tzinfo.
 *  Failure: FALSE.
 */
BOOL WINAPI SetTimeZoneInformation( const TIME_ZONE_INFORMATION *tzinfo )
{
    NTSTATUS status;
    status = RtlSetTimeZoneInformation( (const RTL_TIME_ZONE_INFORMATION *)tzinfo );
    if ( status != STATUS_SUCCESS )
        SetLastError( RtlNtStatusToDosError(status) );
    return !status;
}

/***********************************************************************
 *              SystemTimeToTzSpecificLocalTime  (KERNEL32.@)
 *
 *  Convert a utc system time to a local time in a given time zone.
 *
 * PARAMS
 *  lpTimeZoneInformation [in]  The desired time zone.
 *  lpUniversalTime       [in]  The utc time to base local time on.
 *  lpLocalTime           [out] The local time in the time zone.
 *
 * RETURNS
 *  Success: TRUE. lpLocalTime contains the converted time
 *  Failure: FALSE.
 */

BOOL WINAPI SystemTimeToTzSpecificLocalTime(
    LPTIME_ZONE_INFORMATION lpTimeZoneInformation,
    LPSYSTEMTIME lpUniversalTime, LPSYSTEMTIME lpLocalTime )
{
    FILETIME ft;
    LONG lBias;
    LONGLONG llTime;
    TIME_ZONE_INFORMATION tzinfo;

    if (lpTimeZoneInformation != NULL)
    {
        memcpy(&tzinfo, lpTimeZoneInformation, sizeof(TIME_ZONE_INFORMATION));
    }
    else
    {
        if (GetTimeZoneInformation(&tzinfo) == TIME_ZONE_ID_INVALID)
            return FALSE;
    }

    if (!SystemTimeToFileTime(lpUniversalTime, &ft))
        return FALSE;
    FILETIME2LL( &ft, llTime)
    if (!TIME_GetTimezoneBias(&tzinfo, &ft, FALSE, &lBias))
        return FALSE;
    /* convert minutes to 100-nanoseconds-ticks */
    llTime -= (LONGLONG)lBias * 600000000;
    LL2FILETIME( llTime, &ft)

    return FileTimeToSystemTime(&ft, lpLocalTime);
}


/***********************************************************************
 *              TzSpecificLocalTimeToSystemTime  (KERNEL32.@)
 *
 *  Converts a local time to a time in utc.
 *
 * PARAMS
 *  lpTimeZoneInformation [in]  The desired time zone.
 *  lpLocalTime           [in]  The local time.
 *  lpUniversalTime       [out] The calculated utc time.
 *
 * RETURNS
 *  Success: TRUE. lpUniversalTime contains the converted time.
 *  Failure: FALSE.
 */
BOOL WINAPI TzSpecificLocalTimeToSystemTime(
    LPTIME_ZONE_INFORMATION lpTimeZoneInformation,
    LPSYSTEMTIME lpLocalTime, LPSYSTEMTIME lpUniversalTime)
{
    FILETIME ft;
    LONG lBias;
    LONGLONG t;
    TIME_ZONE_INFORMATION tzinfo;

    if (lpTimeZoneInformation != NULL)
    {
        memcpy(&tzinfo, lpTimeZoneInformation, sizeof(TIME_ZONE_INFORMATION));
    }
    else
    {
        if (GetTimeZoneInformation(&tzinfo) == TIME_ZONE_ID_INVALID)
            return FALSE;
    }

    if (!SystemTimeToFileTime(lpLocalTime, &ft))
        return FALSE;
    FILETIME2LL( &ft, t)
    if (!TIME_GetTimezoneBias(&tzinfo, &ft, TRUE, &lBias))
        return FALSE;
    /* convert minutes to 100-nanoseconds-ticks */
    t += (LONGLONG)lBias * 600000000;
    LL2FILETIME( t, &ft)
    return FileTimeToSystemTime(&ft, lpUniversalTime);
}


/***********************************************************************
 *              GetSystemTimeAsFileTime  (KERNEL32.@)
 *
 *  Get the current time in utc format.
 *
 *  RETURNS
 *   Nothing.
 */
VOID WINAPI GetSystemTimeAsFileTime(
    LPFILETIME time) /* [out] Destination for the current utc time */
{
    LARGE_INTEGER t;
    NtQuerySystemTime( &t );
    time->dwLowDateTime = t.u.LowPart;
    time->dwHighDateTime = t.u.HighPart;
}


/*********************************************************************
 *      TIME_ClockTimeToFileTime    (olorin@fandra.org, 20-Sep-1998)
 *
 *  Used by GetProcessTimes to convert clock_t into FILETIME.
 *
 *      Differences to UnixTimeToFileTime:
 *          1) Divided by CLK_TCK
 *          2) Time is relative. There is no 'starting date', so there is
 *             no need for offset correction, like in UnixTimeToFileTime
 */
#ifndef CLK_TCK
# define CLK_TCK CLOCKS_PER_SEC
#endif
static void TIME_ClockTimeToFileTime(clock_t unix_time, LPFILETIME filetime)
{
    ULONGLONG secs = RtlEnlargedUnsignedMultiply( unix_time, 10000000 );
    secs = RtlExtendedLargeIntegerDivide( secs, CLK_TCK, NULL );
    filetime->dwLowDateTime  = (DWORD)secs;
    filetime->dwHighDateTime = (DWORD)(secs >> 32);
}

/*********************************************************************
 *	GetProcessTimes				(KERNEL32.@)
 *
 *  Get the user and kernel execution times of a process,
 *  along with the creation and exit times if known.
 *
 * PARAMS
 *  hprocess       [in]  The process to be queried.
 *  lpCreationTime [out] The creation time of the process.
 *  lpExitTime     [out] The exit time of the process if exited.
 *  lpKernelTime   [out] The time spent in kernel routines in 100's of nanoseconds.
 *  lpUserTime     [out] The time spent in user routines in 100's of nanoseconds.
 *
 * RETURNS
 *  TRUE.
 *
 * NOTES
 *  olorin@fandra.org:
 *  Would be nice to subtract the cpu time used by Wine at startup.
 *  Also, there is a need to separate times used by different applications.
 *
 * BUGS
 *  KernelTime and UserTime are always for the current process
 */
BOOL WINAPI GetProcessTimes( HANDLE hprocess, LPFILETIME lpCreationTime,
    LPFILETIME lpExitTime, LPFILETIME lpKernelTime, LPFILETIME lpUserTime )
{
    struct tms tms;
    KERNEL_USER_TIMES pti;

    times(&tms);
    TIME_ClockTimeToFileTime(tms.tms_utime,lpUserTime);
    TIME_ClockTimeToFileTime(tms.tms_stime,lpKernelTime);
    if (NtQueryInformationProcess( hprocess, ProcessTimes, &pti, sizeof(pti), NULL))
        return FALSE;
    LL2FILETIME( pti.CreateTime.QuadPart, lpCreationTime);
    LL2FILETIME( pti.ExitTime.QuadPart, lpExitTime);
    return TRUE;
}

/*********************************************************************
 *	GetCalendarInfoA				(KERNEL32.@)
 *
 */
int WINAPI GetCalendarInfoA(LCID lcid, CALID Calendar, CALTYPE CalType,
			    LPSTR lpCalData, int cchData, LPDWORD lpValue)
{
    int ret;
    LPWSTR lpCalDataW = NULL;

    if (NLS_IsUnicodeOnlyLcid(lcid))
    {
      SetLastError(ERROR_INVALID_PARAMETER);
      return 0;
    }

    if (cchData &&
        !(lpCalDataW = HeapAlloc(GetProcessHeap(), 0, cchData*sizeof(WCHAR))))
      return 0;

    ret = GetCalendarInfoW(lcid, Calendar, CalType, lpCalDataW, cchData, lpValue);
    if(ret && lpCalDataW && lpCalData)
      WideCharToMultiByte(CP_ACP, 0, lpCalDataW, cchData, lpCalData, cchData, NULL, NULL);
    HeapFree(GetProcessHeap(), 0, lpCalDataW);

    return ret;
}

/*********************************************************************
 *	GetCalendarInfoW				(KERNEL32.@)
 *
 */
int WINAPI GetCalendarInfoW(LCID Locale, CALID Calendar, CALTYPE CalType,
			    LPWSTR lpCalData, int cchData, LPDWORD lpValue)
{
    if (CalType & CAL_NOUSEROVERRIDE)
	FIXME("flag CAL_NOUSEROVERRIDE used, not fully implemented\n");
    if (CalType & CAL_USE_CP_ACP)
	FIXME("flag CAL_USE_CP_ACP used, not fully implemented\n");

    if (CalType & CAL_RETURN_NUMBER) {
	if (lpCalData != NULL)
	    WARN("lpCalData not NULL (%p) when it should!\n", lpCalData);
	if (cchData != 0)
	    WARN("cchData not 0 (%d) when it should!\n", cchData);
    } else {
	if (lpValue != NULL)
	    WARN("lpValue not NULL (%p) when it should!\n", lpValue);
    }

    /* FIXME: No verification is made yet wrt Locale
     * for the CALTYPES not requiring GetLocaleInfoA */
    switch (CalType & ~(CAL_NOUSEROVERRIDE|CAL_RETURN_NUMBER|CAL_USE_CP_ACP)) {
	case CAL_ICALINTVALUE:
            FIXME("Unimplemented caltype %d\n", CalType & 0xffff);
	    return E_FAIL;
	case CAL_SCALNAME:
            FIXME("Unimplemented caltype %d\n", CalType & 0xffff);
	    return E_FAIL;
	case CAL_IYEAROFFSETRANGE:
            FIXME("Unimplemented caltype %d\n", CalType & 0xffff);
	    return E_FAIL;
	case CAL_SERASTRING:
            FIXME("Unimplemented caltype %d\n", CalType & 0xffff);
	    return E_FAIL;
	case CAL_SSHORTDATE:
	    return GetLocaleInfoW(Locale, LOCALE_SSHORTDATE, lpCalData, cchData);
	case CAL_SLONGDATE:
	    return GetLocaleInfoW(Locale, LOCALE_SLONGDATE, lpCalData, cchData);
	case CAL_SDAYNAME1:
	    return GetLocaleInfoW(Locale, LOCALE_SDAYNAME1, lpCalData, cchData);
	case CAL_SDAYNAME2:
	    return GetLocaleInfoW(Locale, LOCALE_SDAYNAME2, lpCalData, cchData);
	case CAL_SDAYNAME3:
	    return GetLocaleInfoW(Locale, LOCALE_SDAYNAME3, lpCalData, cchData);
	case CAL_SDAYNAME4:
	    return GetLocaleInfoW(Locale, LOCALE_SDAYNAME4, lpCalData, cchData);
	case CAL_SDAYNAME5:
	    return GetLocaleInfoW(Locale, LOCALE_SDAYNAME5, lpCalData, cchData);
	case CAL_SDAYNAME6:
	    return GetLocaleInfoW(Locale, LOCALE_SDAYNAME6, lpCalData, cchData);
	case CAL_SDAYNAME7:
	    return GetLocaleInfoW(Locale, LOCALE_SDAYNAME7, lpCalData, cchData);
	case CAL_SABBREVDAYNAME1:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVDAYNAME1, lpCalData, cchData);
	case CAL_SABBREVDAYNAME2:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVDAYNAME2, lpCalData, cchData);
	case CAL_SABBREVDAYNAME3:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVDAYNAME3, lpCalData, cchData);
	case CAL_SABBREVDAYNAME4:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVDAYNAME4, lpCalData, cchData);
	case CAL_SABBREVDAYNAME5:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVDAYNAME5, lpCalData, cchData);
	case CAL_SABBREVDAYNAME6:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVDAYNAME6, lpCalData, cchData);
	case CAL_SABBREVDAYNAME7:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVDAYNAME7, lpCalData, cchData);
	case CAL_SMONTHNAME1:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME1, lpCalData, cchData);
	case CAL_SMONTHNAME2:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME2, lpCalData, cchData);
	case CAL_SMONTHNAME3:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME3, lpCalData, cchData);
	case CAL_SMONTHNAME4:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME4, lpCalData, cchData);
	case CAL_SMONTHNAME5:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME5, lpCalData, cchData);
	case CAL_SMONTHNAME6:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME6, lpCalData, cchData);
	case CAL_SMONTHNAME7:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME7, lpCalData, cchData);
	case CAL_SMONTHNAME8:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME8, lpCalData, cchData);
	case CAL_SMONTHNAME9:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME9, lpCalData, cchData);
	case CAL_SMONTHNAME10:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME10, lpCalData, cchData);
	case CAL_SMONTHNAME11:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME11, lpCalData, cchData);
	case CAL_SMONTHNAME12:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME12, lpCalData, cchData);
	case CAL_SMONTHNAME13:
	    return GetLocaleInfoW(Locale, LOCALE_SMONTHNAME13, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME1:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME1, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME2:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME2, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME3:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME3, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME4:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME4, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME5:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME5, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME6:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME6, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME7:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME7, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME8:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME8, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME9:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME9, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME10:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME10, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME11:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME11, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME12:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME12, lpCalData, cchData);
	case CAL_SABBREVMONTHNAME13:
	    return GetLocaleInfoW(Locale, LOCALE_SABBREVMONTHNAME13, lpCalData, cchData);
	case CAL_SYEARMONTH:
	    return GetLocaleInfoW(Locale, LOCALE_SYEARMONTH, lpCalData, cchData);
	case CAL_ITWODIGITYEARMAX:
	    if (lpValue) *lpValue = CALINFO_MAX_YEAR;
	    break;
	default: MESSAGE("Unknown caltype %d\n",CalType & 0xffff);
		 return E_FAIL;
    }
    return 0;
}

/*********************************************************************
 *	SetCalendarInfoA				(KERNEL32.@)
 *
 */
int WINAPI	SetCalendarInfoA(LCID Locale, CALID Calendar, CALTYPE CalType, LPCSTR lpCalData)
{
    FIXME("(%08x,%08x,%08x,%s): stub\n",
	  Locale, Calendar, CalType, debugstr_a(lpCalData));
    return 0;
}

/*********************************************************************
 *	SetCalendarInfoW				(KERNEL32.@)
 *
 *
 */
int WINAPI	SetCalendarInfoW(LCID Locale, CALID Calendar, CALTYPE CalType, LPCWSTR lpCalData)
{
    FIXME("(%08x,%08x,%08x,%s): stub\n",
	  Locale, Calendar, CalType, debugstr_w(lpCalData));
    return 0;
}

/*********************************************************************
 *      LocalFileTimeToFileTime                         (KERNEL32.@)
 */
BOOL WINAPI LocalFileTimeToFileTime( const FILETIME *localft, LPFILETIME utcft )
{
    NTSTATUS status;
    LARGE_INTEGER local, utc;

    local.u.LowPart = localft->dwLowDateTime;
    local.u.HighPart = localft->dwHighDateTime;
    if (!(status = RtlLocalTimeToSystemTime( &local, &utc )))
    {
        utcft->dwLowDateTime = utc.u.LowPart;
        utcft->dwHighDateTime = utc.u.HighPart;
    }
    else SetLastError( RtlNtStatusToDosError(status) );

    return !status;
}

/*********************************************************************
 *      FileTimeToLocalFileTime                         (KERNEL32.@)
 */
BOOL WINAPI FileTimeToLocalFileTime( const FILETIME *utcft, LPFILETIME localft )
{
    NTSTATUS status;
    LARGE_INTEGER local, utc;

    utc.u.LowPart = utcft->dwLowDateTime;
    utc.u.HighPart = utcft->dwHighDateTime;
    if (!(status = RtlSystemTimeToLocalTime( &utc, &local )))
    {
        localft->dwLowDateTime = local.u.LowPart;
        localft->dwHighDateTime = local.u.HighPart;
    }
    else SetLastError( RtlNtStatusToDosError(status) );

    return !status;
}

/*********************************************************************
 *      FileTimeToSystemTime                            (KERNEL32.@)
 */
BOOL WINAPI FileTimeToSystemTime( const FILETIME *ft, LPSYSTEMTIME syst )
{
    TIME_FIELDS tf;
    LARGE_INTEGER t;

    t.u.LowPart = ft->dwLowDateTime;
    t.u.HighPart = ft->dwHighDateTime;
    RtlTimeToTimeFields(&t, &tf);

    syst->wYear = tf.Year;
    syst->wMonth = tf.Month;
    syst->wDay = tf.Day;
    syst->wHour = tf.Hour;
    syst->wMinute = tf.Minute;
    syst->wSecond = tf.Second;
    syst->wMilliseconds = tf.Milliseconds;
    syst->wDayOfWeek = tf.Weekday;
    return TRUE;
}

/*********************************************************************
 *      SystemTimeToFileTime                            (KERNEL32.@)
 */
BOOL WINAPI SystemTimeToFileTime( const SYSTEMTIME *syst, LPFILETIME ft )
{
    TIME_FIELDS tf;
    LARGE_INTEGER t;

    tf.Year = syst->wYear;
    tf.Month = syst->wMonth;
    tf.Day = syst->wDay;
    tf.Hour = syst->wHour;
    tf.Minute = syst->wMinute;
    tf.Second = syst->wSecond;
    tf.Milliseconds = syst->wMilliseconds;

    if( !RtlTimeFieldsToTime(&tf, &t)) {
        SetLastError( ERROR_INVALID_PARAMETER);
        return FALSE;
    }
    ft->dwLowDateTime = t.u.LowPart;
    ft->dwHighDateTime = t.u.HighPart;
    return TRUE;
}

/*********************************************************************
 *      CompareFileTime                                 (KERNEL32.@)
 *
 * Compare two FILETIME's to each other.
 *
 * PARAMS
 *  x [I] First time
 *  y [I] time to compare to x
 *
 * RETURNS
 *  -1, 0, or 1 indicating that x is less than, equal to, or greater
 *  than y respectively.
 */
INT WINAPI CompareFileTime( const FILETIME *x, const FILETIME *y )
{
    if (!x || !y) return -1;

    if (x->dwHighDateTime > y->dwHighDateTime)
        return 1;
    if (x->dwHighDateTime < y->dwHighDateTime)
        return -1;
    if (x->dwLowDateTime > y->dwLowDateTime)
        return 1;
    if (x->dwLowDateTime < y->dwLowDateTime)
        return -1;
    return 0;
}

/*********************************************************************
 *      GetLocalTime                                    (KERNEL32.@)
 *
 * Get the current local time.
 *
 * PARAMS
 *  systime [O] Destination for current time.
 *
 * RETURNS
 *  Nothing.
 */
VOID WINAPI GetLocalTime(LPSYSTEMTIME systime)
{
    FILETIME lft;
    LARGE_INTEGER ft, ft2;

    NtQuerySystemTime(&ft);
    RtlSystemTimeToLocalTime(&ft, &ft2);
    lft.dwLowDateTime = ft2.u.LowPart;
    lft.dwHighDateTime = ft2.u.HighPart;
    FileTimeToSystemTime(&lft, systime);
}

/*********************************************************************
 *      GetSystemTime                                   (KERNEL32.@)
 *
 * Get the current system time.
 *
 * PARAMS
 *  systime [O] Destination for current time.
 *
 * RETURNS
 *  Nothing.
 */
VOID WINAPI GetSystemTime(LPSYSTEMTIME systime)
{
    FILETIME ft;
    LARGE_INTEGER t;

    NtQuerySystemTime(&t);
    ft.dwLowDateTime = t.u.LowPart;
    ft.dwHighDateTime = t.u.HighPart;
    FileTimeToSystemTime(&ft, systime);
}

/*********************************************************************
 *      GetDaylightFlag                                   (KERNEL32.@)
 *
 *  Specifies if daylight saving time is in operation.
 *
 * NOTES
 *  This function is called from the Win98's control applet timedate.cpl.
 *
 * RETURNS
 *  TRUE if daylight saving time is in operation.
 *  FALSE otherwise.
 */
BOOL WINAPI GetDaylightFlag(void)
{
    TIME_ZONE_INFORMATION tzinfo;
    return GetTimeZoneInformation( &tzinfo) == TIME_ZONE_ID_DAYLIGHT;
}

/***********************************************************************
 *           DosDateTimeToFileTime   (KERNEL32.@)
 */
BOOL WINAPI DosDateTimeToFileTime( WORD fatdate, WORD fattime, LPFILETIME ft)
{
    struct tm newtm;
#ifndef HAVE_TIMEGM
    struct tm *gtm;
    time_t time1, time2;
#endif

    newtm.tm_sec  = (fattime & 0x1f) * 2;
    newtm.tm_min  = (fattime >> 5) & 0x3f;
    newtm.tm_hour = (fattime >> 11);
    newtm.tm_mday = (fatdate & 0x1f);
    newtm.tm_mon  = ((fatdate >> 5) & 0x0f) - 1;
    newtm.tm_year = (fatdate >> 9) + 80;
    newtm.tm_isdst = -1;
#ifdef HAVE_TIMEGM
    RtlSecondsSince1970ToTime( timegm(&newtm), (LARGE_INTEGER *)ft );
#else
    time1 = mktime(&newtm);
    gtm = gmtime(&time1);
    time2 = mktime(gtm);
    RtlSecondsSince1970ToTime( 2*time1-time2, (LARGE_INTEGER *)ft );
#endif
    return TRUE;
}


/***********************************************************************
 *           FileTimeToDosDateTime   (KERNEL32.@)
 */
BOOL WINAPI FileTimeToDosDateTime( const FILETIME *ft, LPWORD fatdate,
                                     LPWORD fattime )
{
    LARGE_INTEGER       li;
    ULONG               t;
    time_t              unixtime;
    struct tm*          tm;

    li.u.LowPart = ft->dwLowDateTime;
    li.u.HighPart = ft->dwHighDateTime;
    RtlTimeToSecondsSince1970( &li, &t );
    unixtime = t;
    tm = gmtime( &unixtime );
    if (fattime)
        *fattime = (tm->tm_hour << 11) + (tm->tm_min << 5) + (tm->tm_sec / 2);
    if (fatdate)
        *fatdate = ((tm->tm_year - 80) << 9) + ((tm->tm_mon + 1) << 5)
                   + tm->tm_mday;
    return TRUE;
}
